import numpy as np

from quadpy.sphere import Lebedev
from quadpy.e1r import GaussLaguerre

from .quad import spherical_shells, ZERO_ORIGIN


def tent(rnorm, t=3):
    return (1 - rnorm)**t

def bell(rnorm, a=4, b=3):
    return (a * tent(rnorm, b) - b * tent(rnorm, b)) / (a-b)


def GaussLaguerreLebedev(shells, cutoff, scale=0.9, alpha=2):
    """ returns the quadrature described in the original paper
    """

    nradial = len(shells)

    radial = GaussLaguerre(nradial, alpha)

    norm = cutoff * scale / radial.points.max()

    radial.weights *= np.exp(radial.points) * norm**3
    radial.points *= norm

    quad = spherical_shells(radial, [ZERO_ORIGIN if shell == 1 else
                                     Lebedev("%d"  % shell if shell > 4
                                             else "3a") for shell in shells])
    quad.weights *= 4 * np.pi

    return quad


def minmaxscaler(arr):

    lo, hi = arr.min(), arr.max()
    return (arr - lo) / (hi - lo)


def blurring_gaussians(alpha, scale=1, shift=0.5):
    """
    """

    def sigma(pos, masses, radii, cutoff):
        if scale is None:
            r_ = radii
        else:
            r_ = minmaxscaler(radii) * scale + shift

        r = np.linalg.norm(pos, axis=1)
        return r_ * (1 + alpha * r)

    return sigma


def weighed_kernel(kernel=tent, scale=1, shift=0.5):

    def intensity(pos, masses, radii, cutoff):

        # normalize masses to [shift, shift+scale] range
        m = minmaxscaler(masses) * scale + shift

        r = np.linalg.norm(pos, axis=1)

        return m * kernel(r/cutoff)

    return intensity



import numpy as np

from ase.neighborlist import NeighborList
from ase.data import covalent_radii as rcov
from .rotation import CanonicalRotation


class Fingerprint(np.ndarray):
    """
    """

    def __new__(cls, arr, generator):

        obj = np.asarray(arr).view(cls)
        obj.generator = generator

        return obj


    def __sub__(self, other):

        w = self.generator.w_integral
        return np.linalg.norm((np.asarray(self) - np.asarray(other)) * w)


    def plot(self, ax=None, mode=0, **kwargs):

        if ax is None:
            import matplotlib.pyplot as plt
            ax = plt.gca()

        if mode == 0:
            ax.plot(self, **kwargs)
            ax.set_xlabel("Sensor #")
            ax.set_ylabel("Intensity")
            ax.set_yticks(())
        if mode == 1:
            rs = np.linalg.norm(self.generator.sensors, axis=1)
            ax.scatter(rs, self, **kwargs)
            ax.set_xlabel("r, $\mathrm{\AA}$")
            ax.set_ylabel("Intensity")
            ax.set_yticks(())

        return ax


    def plotshell(self, r, dr, ax=None, **kwargs):

        if ax is None:
            import matplotlib.pyplot as plt
            ax = plt.gca()

        rs = np.linalg.norm(self.generator.sensors, axis=1)
        mask = (rs >= r - dr) & (rs <= r + dr)

        if not np.any(mask):
            raise Exception("No sensors in given r ± dr range.")

        r = rs[mask]
        p = self.generator.sensors[mask]

        xy = np.linalg.norm(p[:, :2], axis=1)
        elevation = np.arctan2(xy, p[:, 2])
        azimuth = np.arctan2(p[:, 1], p[:, 0])

        if 'c' in kwargs:
            c = kwargs.pop('c')[mask]
        else:
            c = self[mask]

        im = ax.scatter(azimuth, elevation, c=c, **kwargs)

        ax.set_xlabel('Azimuth')
        ax.set_xlim(-np.pi - 0.3, np.pi + 0.3)
        ax.set_xticks((-np.pi, -np.pi/2, 0, np.pi/2, np.pi))
        ax.set_xticklabels(('-π', '-π/2', '0', 'π/2', 'π'))

        ax.set_ylabel('Elevation')
        ax.set_ylim(-0.3, np.pi + 0.3)
        ax.set_yticks((0, np.pi/2, np.pi))
        ax.set_yticklabels(('0', 'π/2', 'π'))

        return ax, im


class FPGenerator():

    def __init__(self, quadrature, cutoff, atoms=None, I=5E-2, sigma=2,
                 g=None, w=1, debug=False):

        self.g = g
        self.I = I
        self.sigma = sigma
        self.cutoff = cutoff
        self.debug = debug

        self.quad = quadrature
        self.weights = self.quad.weights
        self.sensors = self.quad.points

        if callable(w):
            self.w_integral = w(self.sensors, self.cutoff)
        elif w is 1:
            self.w_integral = np.ones_like(self.weights)
        else:
            self.w_integral = w

        self.atoms = atoms
        if atoms is not None:
            self.nlist = NeighborList([cutoff * 0.5] * len(atoms), skin=0,
                                      self_interaction=False, bothways=True)

            self.nlist.update(atoms)

    def materialize(self, atoms):
        return FPGenerator(self.quad, self.cutoff, atoms=atoms, I=self.I,
                           sigma=self.sigma, g=self.g, w=self.w_integral)

    def __getitem__(self, iatom):
        # TODO let it work with slices

        if self.atoms == None or len(self.atoms) <= iatom:
            raise Exception("Check if atoms object exist or has %d atoms." % 
                            (iatom + 1))

        # get all the atoms in the environment, offsets keep the periodicity
        # information
        neighbors, offsets = self.nlist.get_neighbors(iatom)

        # positions (relative to the central atom), covalent radii and
        # atomic masses will be used to calculate the "importance" of atoms
        # in kernel minisum; intensity and spread of the gaussians associated
        # with each atom

        # calculate atomic positions relative to the atom of interest
        pos = (self.atoms.positions[neighbors] +
               offsets @ self.atoms.cell) - self.atoms[iatom].position

        # get the covalent radii
        radii = rcov[self.atoms.numbers[neighbors]]

        # and masses
        masses = self.atoms.get_masses()[neighbors]

        # calculate g
        if callable(self.g):
            g = self.g(pos, masses, radii, self.cutoff)
        else:
            g = masses * np.exp(-np.linalg.norm(pos, axis=1))

        # evaluate sigma for each atom; if not callable it is 
        # sigma * covalent radius of the atom
        if callable(self.sigma):
            sigma = self.sigma(pos, masses, radii, self.cutoff)
        else:
            sigma = self.sigma * radii

        # evaluate I for each atom; if not callable it is 
        # I * mass of the atom
        if callable(self.I):
            I = self.I(pos, masses, radii, self.cutoff)
        else:
            I = self.I * masses

        #   set stage for minisum calculation to find the canonical coordinates 
        rotator = CanonicalRotation(pos, g, debug=self.debug)

        #   rotate sensors
        sensors = self.sensors @ rotator.Rinv

        if self.debug:
            from ase.io import write
            from ase.atoms import Atoms

            write("%s_%d.xyz" % (self.atoms.get_chemical_formula(), iatom),
                  Atoms(self.atoms.numbers[neighbors],
                        positions=pos @ rotator.R) +
                  Atoms([2]*len(sensors), positions=self.sensors))


        # calculate (R_atom - r_sensor)^2
        r = np.linalg.norm(sensors[:, None] - pos, axis=2)

        # return a vector vec length len(sensors) where vec[j] is:
        # \Sum_{i} w_j * (I * exp(-(R_i - r_j)^2/sigma))
        # i runs over atoms, R is the position of an atom, r is the positon of
        # a sensor.
        # I and sigma are functions of atomic mass and covalent radius
        return Fingerprint(((I * np.exp(-0.5 * (r/sigma)**2) / sigma) *
                             self.weights[:, None]).sum(axis=1), self)



"""
kernel minisum
"""

import numpy as np

from math import sin, cos, sqrt
from quadpy.sphere import Lebedev

from .kernel import ExponentiatedCosine as EC


class CanonicalRotation():


    def __init__(self, x, g, kernel=EC, maxiter=100, convthr=1E-14,
                 n3d=31, n2d=30, debug=False):

        self.x = x
        self.g = g
        self.kernel = kernel(x, g)

        self.n2d = n2d
        self.n3d = n3d

        self.maxiter = maxiter
        self.convthr = convthr
        self.debug = debug

        self.calculate_basis()


    def descent(self, w, normal=None):

        if self.debug:
            print("  i   change   len")
            print("%3d  ******** %5.2f" % (0, self.score(w)[0]))

        converged = False
        old_w = old_dw = None

        for i in range(self.maxiter):

            old_w, (w, old_dw) = w, self.kernel.gradient(w, old_w, old_dw,
                                                         normal)

            diff = np.linalg.norm(w - old_w)

            if self.debug:
                print("%3d %9.2E %5.2f" % (i+1, diff, self.score(w)[0]))

            if diff < self.convthr:
                converged = True
                break

        if converged:
            return w, self.kernel.score(w)[0]
        else:
            return None, 1E15


    def run_minisum(self, normal=None):

        if normal is None:
            ws = Lebedev("%d"%self.n3d).points
        else:
            if normal[0] ** 2 + normal[1] ** 2 > 1E-15:
                v = np.array((-normal[1], normal[0], 0))
            else:
                v = np.array((normal[2], 0, -normal[0]))

            v /= np.linalg.norm(v)
            u = np.cross(normal, v)

            ws = [v * sin(theta) + u * cos(theta) for theta in
                      np.linspace(-np.pi, np.pi, self.n2d)]

        vec = None
        best = 1E15
        for w in ws:
            tmp, current = self.descent(w, normal=normal)

            if current < best:
                vec = tmp
                best = current

        if vec is None:
            raise Exception

        return vec


    def calculate_basis(self):

        u = self.run_minisum()
        v = self.run_minisum(normal=u)
        w = np.cross(u, v)


        front = (w @ self.x.T) > 0
        probe = (u+v) * sqrt(0.5)

        score1 = self.kernel.score(probe, mask=front)
        score2 = self.kernel.score(probe, mask=~front)

        self.basis = u, v, w if score1 <= score2 else -w

        return self.basis


    def rotated(self):

        return self.x @ self.R


    @property
    def R(self):
        return self.Rinv.T


    @property
    def Rinv(self):
        return np.vstack(self.basis)



"""
The kernels for kernel minisum approach. see the original paper for the
description of the implemented ExponentiatedCosine and SquareAngle kernels.

If a new kernel needed, it should inherit Kernel class, and provide the
following functional signatures:
    score(w, mask=None) -> [scores]
    gradient(w, old_w=None, old_dw=None, normal=None) -> w, dw

"""

import numpy as np

class Kernel():

    def __init__(self, x, g):

        self.x = x
        self.g = g
        self.e = self.x / np.linalg.norm(self.x, axis=1, keepdims=True)


    def _preprocess_args(w, mask, N):

        return (w[None, :] if len(w.shape) == 1 else w,
                np.ones(N, dtype=bool) if mask is None else mask)


    def _get_alpha(w, dw, old_w, old_dw):

        if old_w is None:
            return 0.1

        ddw = dw - old_dw
        return abs((w - old_w) @ ddw / np.linalg.norm(ddw) ** 2)


class ExponentiatedCosine(Kernel):

    def score(self, w, mask=None):

        w, mask = Kernel._preprocess_args(w, mask, len(self.g))

        return (self.g[mask] *
                np.exp(-w @ self.e[mask].T + 1) - 1).sum(axis=1)

    def gradient(self, w, old_w=None, old_dw=None, normal=None):

        dw = np.sum(-(self.g * np.exp(-w @ self.e.T))[:, None] * self.e, axis=0)
        dw -= (dw @ w) * w

        # restrict the search if normal is provided
        if normal is not None:
            dw -= normal * (dw @ normal)

        wnew = w - Kernel._get_alpha(w, dw, old_w, old_dw) * dw

        # fix possible rounding errors in restricted case
        if normal is not None:
            wnew -= normal * (wnew @ normal)

        return wnew / np.linalg.norm(wnew), dw


class SquareAngle(Kernel):


    def score(self, w, mask=None):

        w, mask = Kernel._preprocess_args(w, mask, len(self.g))

        return (self.g[mask] * 0.5 *
                np.arccos(w @ self.e[mask].T) ** 2).sum(axis=1)


    def gradient(self, w, old_w=None, old_dw=None, normal=None):

        wxt = w @ self.e.T

        with np.errstate(divide = 'ignore', invalid = 'ignore'):
            C = np.arccos(wxt) / np.sqrt(1 - wxt ** 2)

        C[wxt >  0.9990] = 1.0
        C[wxt < -0.9848] = 17.0866


        dw = np.sum((-self.g * C)[:, None] * self.e, axis=0)
        dw -= (dw @ w) * w

        # restrict the search if normal is provided
        if normal is not None:
            dw -= normal * (dw @ normal)

        wnew = w - Kernel._get_alpha(w, dw, old_w, old_dw) * dw

        # fix possible rounding errors in restricted case
        if normal is not None:
            wnew -= normal * (wnew @ normal)

        return wnew / np.linalg.norm(wnew), dw



import numpy as np


class Quad:


    def __init__(self, points, weights):

        self.points = points
        self.weights = weights



def compose(*args):

    points = []
    weights = []

    for args in args:

        points.append(args.points)
        weights.append(args.weights)


    return Quad(np.column_stack(points),
                np.column_stack(weights))


def spherical_shells(radial, angular):

    try:
        assert len(angular) == len(radial.points)
    except:
        angular = [angular] * len(radial.points)

    points = []
    weights = []

    for r, w, a in zip(radial.points, radial.weights, angular):

        points.append(a.points * r)
        weights.append(a.weights * w)

    return Quad(np.row_stack(points),
                np.hstack(weights))

ZERO_ORIGIN = Quad(np.array((0, 0, 0)), np.array((1, )))



from distutils.core import setup


setup(
    name="ase-decaf",
    version='0.1.0',
    packages=['decaf'],
    license='GNU General Public License v3.0',
    long_description=open('README.txt').read(),
    author='Sencer Selcuk',
    author_email='sselcuk@bnl.gov'
)
